/*
 * Copyright (c) 2003 David E. Bakken, his research students, and Washington State University.
 * Please see the file LICENSE.pdf for more details on terms of use.
 */

package ZEN_IDL.be;

import ZEN_IDL.ast.*;
import ZEN_IDL.be.symbolTableUtils.*;
import ZEN_IDL.be.helperUtils.*;
import util.*;

import java.io.*;
import java.util.*;

/*********************************************************
* Generates code for exceptions
*
* @author Olav Haugan
*
* 14.10.01
*********************************************************/
public class MQCExceptionGenerator
{
	private String safeName; /* The safe name of the excepton */
	private String packagePath; /* The path of this package */
	private String packageStr; /* The package string for this package */
	private String packageName;/* The short package name */
	private String exceptionName; /* The name of the exception */
	private Vector exTypeSpec; 
	private Vector exDecl;

	//variables to ensure that variables in the generated code
	//aren't declared more than once.
	private boolean doubleDef;
	private boolean longDef;
	private boolean floatDef;
	
	private boolean forceEndian;
	private boolean bigEndian;	
			
	//The configuration options
	private MQCConfig config;
	
	private MQCCodeGenHelper codeGenHelper; /* The code generator helper object */
		
	/*********************************************************
	* Constructor. Initializes variables.
	* 
	*********************************************************/
	public MQCExceptionGenerator(MQCConfig config)
	{
		this.config = config;
		codeGenHelper = new MQCCodeGenHelper(config);		
		safeName = null;
		packageStr = null;
		packagePath = null;
		packageName = null;
		exceptionName = null;
		exTypeSpec = null;
		exDecl = null;
		longDef = false;
		doubleDef = false;
		floatDef = false;

		if(config.isForcingEndianess() == 1)
		{
			forceEndian = true;
			bigEndian = true;
		}
		else if(config.isForcingEndianess() == 0)
		{
			if(config.isHomogeneousHW()) //big endian
			{
				forceEndian = true;
				bigEndian = true;
			}
			else
			{
				forceEndian = false;
				bigEndian = false;
			}
		}
		else
		{
			forceEndian = true;
			bigEndian = false;
		}
	}
	
	/*********************************************************
	* Initializes variables
	* 
	*********************************************************/
	public void init()
	{
		longDef = false;
		doubleDef = false;
		floatDef = false;	
	}

	/*********************************************************
	* Sets the safe name variable
	* 
	* @param String The full name of the interface
	*
	*********************************************************/
	public void setSafeName( String safeName )
	{
		this.safeName = safeName;
	}
	
	/*********************************************************
	* Sets the package path variable
	* 
	* @param String The package path
	*
	*********************************************************/
	public void setPackagePath( String packagePath )
	{
		this.packagePath = packagePath;
	}

	/*********************************************************
	* Sets the package string variable
	* 
	* @param String The package string
	*
	*********************************************************/
	public void setPackageStr( String packageStr )
	{
		this.packageStr = packageStr;
	}

	/*********************************************************
	* Sets the package name
	* 
	* @param String The package name
	*
	*********************************************************/
	public void setPackageName( String packageName )
	{
		this.packageName = packageName;
	}
	
	/*********************************************************
	* Sets the exception name
	* 
	* @param String The exception name
	*
	*********************************************************/
	public void setExceptionName( String exceptionName )
	{
		this.exceptionName = exceptionName;
	}

	/*********************************************************
	* Sets the exception type specification.
	* 
	* @param Vector The exception type spec
	*
	*********************************************************/
	public void setExTypeSpec( Vector exTypeSpec )
	{
		this.exTypeSpec = exTypeSpec;
	}

	/*********************************************************
	* Sets the exception declaration 
	* 
	* @param Vector The exception declaration
	*
	*********************************************************/
	public void setExDecl( Vector exDecl )
	{
		this.exDecl = exDecl;
	}
	
	/*********************************************************
	* Generates the exception class
	*********************************************************/
	public void generateException()
	{
		PrintWriter pw = codeGenHelper.makeFile(packagePath, exceptionName );

		if(packagePath.length() != 0)
			pw.println(packageStr);
		pw.println();

		pw.println("final public class " + exceptionName + " extends mqc.exception.UserException");
		pw.println("{");
//		pw.println("\tpublic static java.lang.String id = " + codeGenHelper.generateID(packagePath, safeName) + ";");
		pw.println();
		printDeclarations(pw);
		pw.println();
		printConstructors(pw);
		pw.println();
		printToString(pw);
		pw.println("};");
		pw.println("// End of file.");
		pw.flush();
		pw.close();
	}
	
	/*********************************************************
	* Generates the holder for an exception.
	*********************************************************/
	public void generateExHolder()
	{
		if(config.isInline())
		{
			generateInlineExHolder();
			return;
		}
			
		PrintWriter pw = codeGenHelper.makeFile(packagePath, exceptionName + "Holder" );

		if(packagePath.length() != 0)
			pw.println(packageStr);
		pw.println();

		pw.println("final public class " + exceptionName + "Holder");
		pw.println("{");
		pw.println("\tpublic " + codeGenHelper.generateFullObjName(packagePath, exceptionName) + " value;");
		pw.println();

		pw.println("\tpublic " + exceptionName + "Holder()" );
		pw.println("\t{");
		pw.println("\t}");		
		pw.println();
		
		pw.println("\tpublic " + exceptionName + "Holder( final " + codeGenHelper.generateFullObjName(packagePath, exceptionName) + " value )" );
		pw.println("\t{");
		pw.println("\t\tthis.value = value");
		pw.println("\t}");		
		pw.println();

		pw.println("\tpublic void _read(final mqc.CDRInputStream _msg)");
		pw.println("\t{");
		pw.println("\t\tvalue = " + exceptionName + "Helper.read(_msg);");
		pw.println("\t}");		
		pw.println();

		pw.println("\tpublic void _write(final mqc.CDROutputStream output)");
		pw.println("\t{");
		pw.println("\t\t" + exceptionName + "Helper.write(output, value);");
		pw.println("\t}");		
		pw.println();
		
		pw.println("};");
		pw.println("// End of file.");
		pw.flush();
		pw.close();
	}

	/*********************************************************
	* Generates the holder for an exception with inline 
	* marshal support.
	*********************************************************/
	public void generateInlineExHolder()
	{
		PrintWriter pw = codeGenHelper.makeFile(packagePath, exceptionName + "Holder" );

		if(packagePath.length() != 0)
			pw.println(packageStr);
		pw.println();

		pw.println("final public class " + exceptionName + "Holder");
		pw.println("{");
		pw.println("\tpublic " + codeGenHelper.generateFullObjName(packagePath, exceptionName) + " value;");
		pw.println();

		pw.println("\tpublic " + exceptionName + "Holder()" );
		pw.println("\t{");
		pw.println("\t}");		
		pw.println();
		
		pw.println("\tpublic " + exceptionName + "Holder( final " + codeGenHelper.generateFullObjName(packagePath, exceptionName) + " value )" );
		pw.println("\t{");
		pw.println("\t\tthis.value = value");
		pw.println("\t}");		
		pw.println();

		pw.println("\tpublic void _read(final byte[] buffer, final int ptr, final boolean bigEndian)");
		pw.println("\t{");
		pw.println("\t\tvalue = " + exceptionName + "Helper.read(buffer, ptr, bigEndian);");
		pw.println("\t}");		
		pw.println();

		pw.println("\tpublic void _write(final byte[] buffer)");
		pw.println("\t{");
		pw.println("\t\t" + exceptionName + "Helper.write(buffer, value);");
		pw.println("\t}");		
		pw.println();
		
		pw.println("};");
		pw.println("// End of file.");
		pw.flush();
		pw.close();
	}

	/*********************************************************
	* Generates the helper for an exception.
	*********************************************************/
	public void generateExHelper()
	{
		if(config.isInline())
		{
			generateInlineExHelper();
			return;
		}
			
		PrintWriter pw = codeGenHelper.makeFile(packagePath, exceptionName + "Helper" );

		String fullName = codeGenHelper.generateFullObjName(packagePath, exceptionName);
		
		if(packagePath.length() != 0)
			pw.println(packageStr);
		pw.println();

		pw.println("final public class " + exceptionName + "Helper");
		pw.println("{");

		pw.println("\tpublic static " + fullName + " read(final mqc.CDRInputStream _msg)");
		pw.println("\t{");
		pw.println("\t\tif( !_msg.read_string().equals(id()))");
		pw.println("\t\t{");
		pw.println("\t\t\t//throw new mqc.exception.MARSHAL(\"Mismatched repository id\");");
		pw.println("\t\t}");
		pw.println("\t\tfinal " + fullName + " _result = new " + fullName + "();");		

		Enumeration eType = exTypeSpec.elements();
		Enumeration eDecl = exDecl.elements();		
		StructDeclarator declarator = null;
		String type = null;
		String name = null;

		pw.println("\t\t_result.id = id();");

		//Only works for simple types
		while(eType.hasMoreElements())
		{
			type = codeGenHelper.mapIDL2RWMap(eType.nextElement().toString());
			declarator = (StructDeclarator) eDecl.nextElement();
			name = declarator.printAsDeclaration();
			pw.println("\t\t_result." + name + " = _msg.read_" + type + "();");
		}

		pw.println("\t\treturn _result;");
		pw.println("\t}");		
		pw.println();

		pw.println("\tpublic static void write(final mqc.CDROutputStream output, final " + fullName + " value)");
		pw.println("\t{");
		pw.println("\t\toutput.write_string(id());");

		eType = exTypeSpec.elements();
		eDecl = exDecl.elements();		

		//Only works for simple types
		while(eType.hasMoreElements())
		{
			type = codeGenHelper.mapIDL2RWMap(eType.nextElement().toString());
			declarator = (StructDeclarator) eDecl.nextElement();
			name = declarator.printAsDeclaration();
			pw.println("\t\toutput.write_" + type + "(value." + name + ");");
		}

		pw.println("\t}");		
		pw.println();

		pw.println("\tpublic static java.lang.String id()");
		pw.println("\t{");
		pw.println("\t\treturn \"" + codeGenHelper.generateID(packagePath, safeName) + "\";");		
		pw.println("\t}");
		pw.println("};");
		pw.println("// End of file.");
		pw.flush();
		pw.close();
	}

	/*********************************************************
	* Generates the helper for an exception with inline 
	* marshal support.
	*********************************************************/
	public void generateInlineExHelper()
	{
		PrintWriter pw = codeGenHelper.makeFile(packagePath, exceptionName + "Helper" );

		String fullName = codeGenHelper.generateFullObjName(packagePath, exceptionName);
		
		if(packagePath.length() != 0)
			pw.println(packageStr);
		pw.println();

		pw.println("final public class " + exceptionName + "Helper");
		pw.println("{");

		pw.print("\tpublic static " + fullName + " read(final byte[] _msg, int ptr, final boolean bigEndian)");
		pw.println(" throws mqc.exception.MARSHAL");
		pw.println("\t{");
		pw.println("\t\tStringBuffer idStr = new StringBuffer();");
		if(config.getClientProtocol().equals("MQCIOP"))
		{
			generateMQCInlineExHelperReader(pw);
		}
		else
		{
			generateInlineExHelperReader(pw);
		}
		
		init(); //reset variables

		//Check if server supports MQC-IOP and another protocol
		Vector protocols = config.getServerProtocol();
		boolean mqcIOP = false;

		Enumeration e = protocols.elements();
		while(e.hasMoreElements())
		{
			if(((String) e.nextElement()).equals("MQCIOP"))
			{
				mqcIOP = true;
				break;
			}
		}	
		boolean moreThanOne = ((protocols.size() > 1) ? true : false);

		pw.println("\tpublic static byte[] write(byte[] _msg, mqc.holders.IntHolder ptr, " + fullName + " value)");
		pw.println("\t{");

		if(mqcIOP && moreThanOne)
		{
			pw.println("\t\t/* MQC */");
			pw.println("\t\tif((_msg[0] == 0x4D) && (_msg[1] == 0x51) && (_msg[2] == 0x43))");
			pw.println("\t\t\treturn writeMQC(_msg, ptr, value);");
			pw.println();
		}
		if(mqcIOP && !moreThanOne)
		{
			generateMQCInlineExHelperWriter(pw);
		}
		else
		{
			generateInlineExHelperWriter(pw);
		}

		init(); //reset variables
		
		if(mqcIOP && moreThanOne)
		{
			pw.println();
			pw.println("\tpublic static byte[] writeMQC(byte[] _msg, mqc.holders.IntHolder ptr, " + fullName + " value)");
			pw.println("\t{");
			generateMQCInlineExHelperWriter(pw);				
		}

		pw.println("\tpublic static java.lang.String id()");
		pw.println("\t{");
		pw.println("\t\treturn \"" + codeGenHelper.generateID(packagePath, safeName) + "\";");		
		pw.println("\t}");

		pw.println("};");
		pw.println("// End of file.");
		pw.flush();
		pw.close();
	}


	/*********************************************************
	* Generates the writer helper for an exception with inline 
	* marshal support.
	*
	* @param PrintWriter The output stream
	*********************************************************/
	public void generateInlineExHelperWriter(PrintWriter pw)
	{
		Enumeration eType = exTypeSpec.elements();
		Enumeration eDecl = exDecl.elements();		
		StructDeclarator declarator = null;
		String type = null;
		String name = null;
		String id = codeGenHelper.generateID(packagePath, safeName);
		int idLen = id.length();

		if(config.isPadding())
		{
			printAlign(pw, "long", true);
		}
		pw.println("\t\t{");
		pw.println("\t\t\tbyte[] tmp = new byte[ptr.value + 4 + " + (idLen + 1) + "];");
		pw.println("\t\t\tSystem.arraycopy(_msg, 0, tmp, 0, _msg.length);");
		pw.println("\t\t\t_msg = tmp;");
		pw.println("\t\t}");
		
		if(forceEndian && !bigEndian)
		{
			pw.println("\t\t_msg[ptr.value++] = (byte) " + ((idLen +1) >> 0) + ";" );
			pw.println("\t\t_msg[ptr.value++] = (byte) " + ((idLen +1)  >> 8) + ";");
			pw.println("\t\t_msg[ptr.value++] = (byte) " + ((idLen +1) >> 16) + ";" );
			pw.println("\t\t_msg[ptr.value++] = (byte) " + ((idLen +1) >> 24)+ ";");							
		}
		else
		{
			pw.println("\t\t_msg[ptr.value++] = (byte) " + ((idLen +1) >> 24) + ";" );
			pw.println("\t\t_msg[ptr.value++] = (byte) " + ((idLen +1)  >> 16) + ";");
			pw.println("\t\t_msg[ptr.value++] = (byte) " + ((idLen +1) >> 8) + ";" );
			pw.println("\t\t_msg[ptr.value++] = (byte) " + (idLen +1) + ";");							
		}
		pw.println("\t\tSystem.arraycopy(id().getBytes(), 0, _msg, ptr.value, id().getBytes().length);");
		pw.println("\t\tptr.value += " + idLen + ";");						
		pw.println("\t\t_msg[ptr.value++] = (byte) 0;");						
		pw.println();
		
		eType = exTypeSpec.elements();
		eDecl = exDecl.elements();		

		//Only works for simple types
		while(eType.hasMoreElements())
		{
			type = eType.nextElement().toString();
			declarator = (StructDeclarator) eDecl.nextElement();
			name = declarator.printAsDeclaration();
			doMarshal(pw, type, name);
		}
		pw.println("\t\treturn _msg;");		
		pw.println("\t}");		
		pw.println();
	}

	/*********************************************************
	* Generates the writer helper for an exception with inline 
	* marshal support for use with MQC-IOP protocol.
	*
	* @param PrintWriter The output stream
	*********************************************************/
	public void generateMQCInlineExHelperWriter(PrintWriter pw)
	{
		Enumeration eType = exTypeSpec.elements();
		Enumeration eDecl = exDecl.elements();		
		StructDeclarator declarator = null;
		String type = null;
		String name = null;
		String id = codeGenHelper.generateID(packagePath, safeName);
		int idLen = id.length();

		int index = Constants.MQC_SIZE_OF_HDR + Constants.MQC_SIZE_OF_REPLY_HDR;
		
		if(config.isPadding())
		{
			index = codeGenHelper.align("long", index);
		}

		if(forceEndian && !bigEndian)
		{
			pw.println("\t\t_msg[" + index++ + "] = (byte) " + ((idLen +1) >> 0) + ";" );
			pw.println("\t\t_msg[" + index++ + "] = (byte) " + ((idLen +1)  >> 8) + ";");
			pw.println("\t\t_msg[" + index++ + "] = (byte) " + ((idLen +1) >> 16) + ";" );
			pw.println("\t\t_msg[" + index++ + "] = (byte) " + ((idLen +1) >> 24)+ ";");							
		}
		else
		{
			pw.println("\t\t_msg[" + index++ + "] = (byte) " + ((idLen +1) >> 24) + ";" );
			pw.println("\t\t_msg[" + index++ + "] = (byte) " + ((idLen +1)  >> 16) + ";");
			pw.println("\t\t_msg[" + index++ + "] = (byte) " + ((idLen +1) >> 8) + ";" );
			pw.println("\t\t_msg[" + index++ + "] = (byte) " + (idLen +1) + ";");							
		}
		pw.println("\t\tSystem.arraycopy(id().getBytes(), 0, _msg, " +  index + ", " + idLen + ");");
		index += idLen;
		pw.println("\t\t_msg[" + index++ + "] = (byte) 0;");						
		pw.println();
		
		eType = exTypeSpec.elements();
		eDecl = exDecl.elements();		

		//Only works for simple types
		while(eType.hasMoreElements())
		{
			type = eType.nextElement().toString();
			declarator = (StructDeclarator) eDecl.nextElement();
			name = declarator.printAsDeclaration();
			index = doMQCMarshal(pw, type, name, index);
		}
		pw.println("\t\tptr.value = " + index + ";");		
		pw.println("\t\treturn _msg;");		
		pw.println("\t}");		
		pw.println();

	}


	/*********************************************************
	* Generates the reader helper for an exception with inline 
	* marshal support.
	*
	* @param PrintWriter The outputstream to print to.
	*********************************************************/
	private void generateInlineExHelperReader(PrintWriter pw)
	{
		String fullName = codeGenHelper.generateFullObjName(packagePath, exceptionName);

		if(config.isPadding())
		{
			printAlign(pw, "long", false);
		}
		pw.println("\t\tint len;");
		if(forceEndian && bigEndian)
		{
			pw.println("\t\t\tlen = (int) (((_msg[ptr++] & 0xFF) << 24) | ((_msg[ptr++] & 0xFF) << 16) | ((_msg[ptr++] & 0xFF) << 8) | (_msg[ptr++] & 0xFF));");
		}
		else if(forceEndian && !bigEndian)
		{
			pw.println("\t\t\tlen = (int) ((_msg[ptr++] & 0xFF) | ((_msg[ptr++] & 0xFF) << 8) | ((_msg[ptr++] & 0xFF) << 16) | ((_msg[ptr++] & 0xFF) << 24));");		
		}	
		else
		{
			pw.println("\t\tif(bigEndian)");
			pw.println("\t\t\tlen = (int) (((_msg[ptr++] & 0xFF) << 24) | ((_msg[ptr++] & 0xFF) << 16) | ((_msg[ptr++] & 0xFF) << 8) | (_msg[ptr++] & 0xFF));");
			pw.println("\t\telse");
			pw.println("\t\t\tlen = (int) ((_msg[ptr++] & 0xFF) | ((_msg[ptr++] & 0xFF) << 8) | ((_msg[ptr++] & 0xFF) << 16) | ((_msg[ptr++] & 0xFF) << 24));");
		}
		pw.println("\t\tfor(int i = 0; i < len-1; i++)");
		pw.println("\t\t{");
		pw.println("\t\t\tidStr.append((char) _msg[ptr++]);");
		pw.println("\t\t}");
		pw.println("\t\tptr++;");
		pw.println("\t\tif( !idStr.toString().equals(id()))");
		pw.println("\t\t{");
		pw.println("\t\t\tthrow new mqc.exception.MARSHAL(\"Mismatched repository id\");");
		pw.println("\t\t}");
		pw.println("\t\tfinal " + fullName + " _result = new " + fullName + "();");		
		pw.println("\t\t_result.id = id();");
		String id = codeGenHelper.generateID(packagePath, safeName);
		int idLen = id.length();
		
		Enumeration eType = exTypeSpec.elements();
		Enumeration eDecl = exDecl.elements();		
		StructDeclarator declarator = null;
		String type = null;
		String name = null;

		//Only works for simple types
		while(eType.hasMoreElements())
		{
			type = eType.nextElement().toString();
			declarator = (StructDeclarator) eDecl.nextElement();
			name = declarator.printAsDeclaration();
			doUnMarshal(pw, type, name);
		}
		
		pw.println("\t\treturn _result;");
		pw.println("\t}");		
		pw.println();
	}
	
	/*********************************************************
	* Generates the reader helper for an exception with inline 
	* marshal support for the MQCIOP protocol.
	*
	* @param PrintWriter The outputstream to print to.
	*********************************************************/
	private void generateMQCInlineExHelperReader(PrintWriter pw)
	{
		String fullName = codeGenHelper.generateFullObjName(packagePath, exceptionName);
		int index = Constants.MQC_SIZE_OF_HDR + Constants.MQC_SIZE_OF_REPLY_HDR;
		
		if(config.isPadding())
		{
			index = codeGenHelper.align("long", index);
		}
		pw.println("\t\tint len;");
		if(forceEndian && bigEndian)
		{
			pw.println("\t\t\tlen = (int) (((_msg[" + index++ + "] & 0xFF) << 24) | ((_msg[" + index++ + "] & 0xFF) << 16) | ((_msg[" + index++ + "] & 0xFF) << 8) | (_msg[" + index++ + "] & 0xFF));");
		}
		else if(forceEndian && !bigEndian)
		{
			pw.println("\t\t\tlen = (int) ((_msg[" + index++ + "] & 0xFF) | ((_msg[" + index++ + "] & 0xFF) << 8) | ((_msg[" + index++ + "] & 0xFF) << 16) | ((_msg[" + index++ + "] & 0xFF) << 24));");		
		}	
		else
		{
			int tmpIndex = index;
			pw.println("\t\tif(bigEndian)");
			pw.println("\t\t\tlen = (int) (((_msg[" + index++ + "] & 0xFF) << 24) | ((_msg[" + index++ + "] & 0xFF) << 16) | ((_msg[" + index++ + "] & 0xFF) << 8) | (_msg[" + index++ + "] & 0xFF));");
			pw.println("\t\telse");
			pw.println("\t\t\tlen = (int) ((_msg[" + tmpIndex++ + "] & 0xFF) | ((_msg[" + tmpIndex++ + "] & 0xFF) << 8) | ((_msg[" + tmpIndex++ + "] & 0xFF) << 16) | ((_msg[" + tmpIndex++ + "] & 0xFF) << 24));");
		}
		pw.println("\t\tint i;");
		pw.println("\t\tfor(i = 0; i < len-1; i++)");
		pw.println("\t\t{");
		pw.println("\t\t\tidStr.append((char) _msg[" + index + "+i]);");
		pw.println("\t\t}");
		pw.println("\t\tptr = " + index + " + i + 1;");
		pw.println("\t\tif( !idStr.toString().equals(id()))");
		pw.println("\t\t{");
		pw.println("\t\t\tthrow new mqc.exception.MARSHAL(\"Mismatched repository id\");");
		pw.println("\t\t}");
		pw.println("\t\tfinal " + fullName + " _result = new " + fullName + "();");		
		pw.println("\t\t_result.id = id();");

		String id = codeGenHelper.generateID(packagePath, safeName);
		int idLen = id.length();
		
		Enumeration eType = exTypeSpec.elements();
		Enumeration eDecl = exDecl.elements();		
		StructDeclarator declarator = null;
		String type = null;
		String name = null;

		//Only works for simple types
		while(eType.hasMoreElements())
		{
			type = eType.nextElement().toString();
			declarator = (StructDeclarator) eDecl.nextElement();
			name = declarator.printAsDeclaration();
			doUnMarshal(pw, type, name);
		}
		
		pw.println("\t\treturn _result;");
		pw.println("\t}");		
		pw.println();
	}

	/*********************************************************
	* Prinsts the member variables declarations
	* 
	* @param PrintWriter The outputstream to print to.
	*
	*********************************************************/
	private void printDeclarations(PrintWriter pw)
	{
		Enumeration eType = exTypeSpec.elements();
		Enumeration eDecl = exDecl.elements();		
		StructDeclarator declarator = null;
		String type = null;
		String name = null;
		while(eType.hasMoreElements())
		{
			type = codeGenHelper.mapIDL2Java(eType.nextElement().toString());
			declarator = (StructDeclarator) eDecl.nextElement();
			name = declarator.printAsDeclaration();
			pw.println("\tpublic " + type + " " + name + ";");
		}
	}

	/*********************************************************
	* Prinsts the constructors
	* 
	* @param PrintWriter The outputstream to print to.
	*
	*********************************************************/
	private void printConstructors(PrintWriter pw)
	{
		pw.println("\tpublic " + safeName + "()");
		pw.println("\t{");
		pw.println("\t}");		

		pw.print("\tpublic " + safeName + "( ");

		Enumeration eType = exTypeSpec.elements();
		Enumeration eDecl = exDecl.elements();		
		StructDeclarator declarator = null;
		String type = null;
		String name = null;
		while(eType.hasMoreElements())
		{
			type = codeGenHelper.mapIDL2Java(eType.nextElement().toString());
			declarator = (StructDeclarator) eDecl.nextElement();
			name = declarator.printAsParameter();
			pw.print(type + " " + name );
			if(eType.hasMoreElements())
				pw.print(", " );
		}
		pw.println(" )" );
		pw.println("\t{");

		eDecl = exDecl.elements();		
		while(eDecl.hasMoreElements())
		{
			declarator = (StructDeclarator) eDecl.nextElement();
			name = declarator.printAsAssignment();
			pw.println("\t\tthis." + name + " = " + name + ";" );
		}

		pw.println("\t}");		
	}

	/*********************************************************
	* Prinsts the toString() function
	* 
	* @param PrintWriter The outputstream to print to.
	*
	*********************************************************/
	private void printToString(PrintWriter pw)
	{
		pw.println("\tpublic String toString()");
		pw.println("\t{");
		pw.println("\t\treturn id;");
		pw.println("\t}");		
	}

	/*********************************************************
	* Prints the inline client marshalling code for basic
	* types.
	*
	* @param PrintWriter The outputstream.
	*        
	* @param String The type
	*        
	* @param String The name
	*********************************************************/
	public void doMarshal(PrintWriter pw, String type, String name )
	{
		if(type.equals("float"))
		{
			if(config.isPadding())
			{
				printAlign(pw, "float", true);
			}
			printBufferInc(pw, type, true );
			doMarshalFloat(pw, name);
		}
		else if(type.equals("double"))
		{
			if(config.isPadding())
			{
				printAlign(pw, "double", true);
			}
			printBufferInc(pw, type, true );
			doMarshalDouble(pw, name);
		}
		else if(type.equals("boolean"))
		{
			pw.print("\t\t_msg[ptr.value++] = (byte) (");
			doMarshalBoolean(pw, name);
		}
		else
		{
			int paramSize = codeGenHelper.getParamSize(type);
			if(config.isPadding() && paramSize > 1)
			{
				printAlign(pw, type, true);
			}
			printBufferInc(pw, type, true );			
			if(forceEndian && !bigEndian)
			{
				for(int i = 0; i < paramSize-1; i++)
				{
					pw.print("\t\t_msg[ptr.value++] = (byte) ");
					pw.println("(value." + name + " >> " + (8 * i) + ");" );
				}
				pw.println("\t\t_msg[ptr.value++] = (byte) (value." + name + " >> " + (8 * (paramSize-1)) + ");" );
			}
			else
			{
				for(int i = paramSize-1; i > 0; i--)
				{
					pw.print("\t\t_msg[ptr.value++] = (byte) ");
					pw.println("(value." + name + " >> " + (8 * i) + ");" );
				}
				pw.println("\t\t_msg[ptr.value++] = (byte) value." + name + ";");
			}			
		}
	}	

	/*********************************************************
	* Prints the inline client marshalling code for basic
	* types for use with MQC-IOP protocol.
	*
	* @param PrintWriter The outputstream.
	*        
	* @param String The type
	*        
	* @param String The name
	*
	* @param int The current index
	*
	* @return int The new index
	*********************************************************/
	public int doMQCMarshal(PrintWriter pw, String type, String name, int index )
	{
		if(type.equals("float"))
		{
			if(config.isPadding())
			{
				index = codeGenHelper.align("float", index);
			}
			index = doMQCMarshalFloat(pw, name, index);
		}
		else if(type.equals("double"))
		{
			if(config.isPadding())
			{
				index = codeGenHelper.align("double", index);
			}
			index = doMQCMarshalDouble(pw, name, index);
		}
		else if(type.equals("boolean"))
		{
			pw.print("\t\t_msg[" + index++ + "] = (byte) (");
			doMarshalBoolean(pw, name);
		}
		else
		{
			int paramSize = codeGenHelper.getParamSize(type);
			if(config.isPadding() && paramSize > 1)
			{
				index = codeGenHelper.align(type, index);
			}
			if(forceEndian && !bigEndian)
			{
				for(int i = 0; i < paramSize-1; i++)
				{
					pw.print("\t\t_msg[" + index++ + "] = (byte) ");
					pw.println("(value." + name + " >> " + (8 * i) + ");" );
				}
				pw.println("\t\t_msg[" + index++ + "] = (byte) (value." + name + " >> " + (8 * (paramSize-1)) + ");" );
			}
			else
			{
				for(int i = paramSize-1; i > 0; i--)
				{
					pw.print("\t\t_msg[" + index++ + "] = (byte) ");
					pw.println("(value." + name + " >> " + (8 * i) + ");" );
				}
				pw.println("\t\t_msg[" + index++ + "] = (byte) value." + name + ";");
			}			
		}
		return index;
	}	

	/*********************************************************
	* Prints the inline client un-marshalling code for basic
	* types in exception.
	*
	* @param PrintWriter The outputstream.
	*        
	* @param String The parameter type
	*        
	* @param String The name of the variable
	*********************************************************/
	public void doUnMarshal(PrintWriter pw, String type, String name)
	{
		if(type.equals("float"))
		{
			if(config.isPadding())
			{
				printAlign(pw, "float", false);
			}
			doUnMarshalFloat(pw, name);
		}
		else if(type.equals("double"))
		{
			if(config.isPadding())
			{
				printAlign(pw, "double", false);
			}
			doUnMarshalDouble(pw, name);
		}
		else if(type.equals("boolean"))
		{
			pw.print("\t\t_result." + name + " = (" + codeGenHelper.mapIDL2Java( type ) + ") (");
			doUnMarshalBoolean(pw);
		}
		else if(type.equals("long long") || type.equals("unsigned long long"))
		{
			if(config.isPadding())
			{
				printAlign(pw, "long long", false);
			}
			doUnMarshalLong(pw, name);
		}
		else
		{
			int paramSize = codeGenHelper.getParamSize(type);
			if(config.isPadding() && paramSize > 1)
			{
				printAlign(pw, type, false);
			}
			if(forceEndian && bigEndian)
			{
				pw.print("\t\t_result." + name + " = (" + codeGenHelper.mapIDL2Java( type ) + ") (");
				for(int i = paramSize-1; i > 0; i--)
				{
					pw.print("((_msg[ptr++] & 0xFF) << " + (8 * i) + ") | ");
				}
				pw.println("(_msg[ptr++] & 0xFF));");
			}
			else if(forceEndian && !bigEndian)
			{
				pw.print("\t\t_result." + name + " = (" + codeGenHelper.mapIDL2Java( type ) + ") (");
				int i = 0;
				for(i = 0; i < paramSize-1; i++)
				{
					pw.print("((_msg[ptr++] & 0xFF) << " + (8 * i) + ") | ");
				}
				pw.println("((_msg[ptr++] & 0xFF) << " + (8*i) + " ));");
			}
			else
			{
				pw.println("\t\tif(bigEndian)");
				pw.println("\t\t{");
				pw.print("\t\t\t_result." + name + " = (" + codeGenHelper.mapIDL2Java( type ) + ") (");
				for(int i = paramSize-1; i > 0; i--)
				{
					pw.print("((_msg[ptr++] & 0xFF) << " + (8 * i) + ") | ");
				}
				pw.println("(_msg[ptr++] & 0xFF));");
				pw.println("\t\t}");
				pw.println("\t\telse");
				pw.println("\t\t{");
				pw.print("\t\t\t_result." + name + " = (" + codeGenHelper.mapIDL2Java( type ) + ") (");
				int i = 0;
				for(i = 0; i < paramSize-1; i++)
				{
					pw.print("((_msg[ptr++] & 0xFF) << " + (8 * i) + ") | ");
				}
				pw.println("((_msg[ptr++] & 0xFF) << " + (8*i) + " ));");
				pw.println("\t\t}");
			}
		}
	}
	
	/*********************************************************
	* Prints alignment code
	* 
	* @param PrintWriter The outputstream
	*
	* @param String parameter type
	*
	* @param boolean true if a holder object, false otherwise
	*********************************************************/
	public void printAlign(PrintWriter pw, String type, boolean holder )
	{
		if(codeGenHelper.getAlignment(type) < 1)
			return;

		if(holder)
		{
			pw.print("\t\tptr.value = (ptr.value + " + (codeGenHelper.getAlignment(type) - 1) + ") & ~" );
			pw.println((codeGenHelper.getAlignment(type)-1) + ";");
		}
		else
		{
			pw.print("\t\tptr = (ptr + " + (codeGenHelper.getAlignment(type) - 1) + ") & ~" );
			pw.println((codeGenHelper.getAlignment(type)-1) + ";");
		}
	}
	
	/*********************************************************
	* Prints the buffer increase code
	* 
	* @param PrintWriter The outputstream
	*
	* @param String parameter type
	*********************************************************/
	public void printBufferInc(PrintWriter pw, String type, boolean client )
	{
		if(codeGenHelper.getParamSize(type) == 0)
			return;

		if(client)
		{
			pw.println("\t\tif(_msg.length < (ptr.value + " + codeGenHelper.getParamSize(type) + "))");
			pw.println("\t\t{");
			pw.println("\t\t\tbyte[] tmp = new byte[_msg.length + " + codeGenHelper.getParamSize(type) + " + 10];");
			pw.println("\t\t\tSystem.arraycopy(_msg, 0, tmp, 0, _msg.length);");
			pw.println("\t\t\t_msg = tmp;");
			pw.println("\t\t}");
		}
		else
		{
			pw.println("\t\tif(_output.length < (ptr.value + " + codeGenHelper.getParamSize(type) + "))");
			pw.println("\t\t{");
			pw.println("\t\t\tbyte[] tmp = new byte[_output.length + " + codeGenHelper.getParamSize(type) + " + 10];");
			pw.println("\t\t\tSystem.arraycopy(_output, 0, tmp, 0, _output.length);");
			pw.println("\t\t\t_output = tmp;");
			pw.println("\t\t}");
		}
	}

	/*********************************************************
	* Prints the marshaling of float in an exception
	* 
	* @param PrintWriter The outputstream.
	*        
	* @param String Name of argument
	*********************************************************/
	private void doMarshalFloat(PrintWriter pw, String name)
	{
		if(!floatDef)
			pw.print("\t\tint fval = ");
		else
			pw.print("\t\tfval = ");
		floatDef = true;
		
		pw.println("java.lang.Float.floatToIntBits(value." + name + ");");		

		int paramSize = codeGenHelper.getParamSize("float");
		if(forceEndian && !bigEndian)
		{
			for(int i = 0; i < paramSize-1; i++)
			{
				pw.print("\t\t_msg[ptr.value++] = (byte) ");
				pw.println("(fval >> " + (8 * i) + ");" );
			}
			pw.println("\t\t_msg[ptr.value++] = (byte) (fval >> " + (8 * (paramSize-1)) + ");" );
		}
		else
		{
			for(int i = paramSize-1; i > 0; i--)
			{
				pw.print("\t\t_msg[ptr.value++] = (byte) ");
				pw.println("(fval >> " + (8 * i) + ");" );
			}
			pw.println("\t\t_msg[ptr.value++] = (byte) fval;");
		}
	}


	/*********************************************************
	* Prints the marshaling of double in an exception
	* 
	* @param PrintWriter The outputstream.
	*        
	* @param String The name of the argument
	*********************************************************/
	private void doMarshalDouble(PrintWriter pw, String name)
	{
		if(!doubleDef)
			pw.print("\t\tlong dval = ");
		else
			pw.print("\t\tdval = ");
		doubleDef = true;
		
		pw.println("java.lang.Double.doubleToLongBits(value." + name + ");");

		int paramSize = codeGenHelper.getParamSize("double");
		if(forceEndian && !bigEndian)
		{
			for(int i = 0; i < paramSize-1; i++)
			{
				pw.print("\t\t_msg[ptr.value++] = (byte) ");
				pw.println("(dval >> " + (8 * i) + ");" );
			}
			pw.println("\t\t_msg[ptr.value++] = (byte) (dval >> " + (8 * (paramSize-1)) + ");" );
		}
		else
		{
			for(int i = paramSize-1; i > 0; i--)
			{
				pw.print("\t\t_msg[ptr.value++] = (byte) ");
				pw.println("(dval >> " + (8 * i) + ");" );
			}
			pw.println("\t\t_msg[ptr.value++] = (byte) dval;");
		}
		
	}

	/*********************************************************
	* Prints the marshaling of boolean in an exception
	* 
	* @param PrintWriter The outputstream.
	*        
	* @param String Name of argument.
	*********************************************************/
	private void doMarshalBoolean(PrintWriter pw, String name)
	{
		pw.println("value." + name + " ? 1 : 0);");
	}

	/*********************************************************
	* Prints the un-marshaling of float in a exception
	* 
	* @param PrintWriter The outputstream.
	*
	* @param String Name of param
	*********************************************************/
	private void doUnMarshalFloat(PrintWriter pw, String name)
	{
		int paramSize = codeGenHelper.getParamSize("float");
		pw.println("\t\tif(bigEndian)");
		pw.println("\t\t{");

		pw.print("\t\t\t_result." + name + " = java.lang.Float.intBitsToFloat((int) (" );
		for(int i = paramSize-1; i > 0; i--)
		{
			pw.print("((_msg[ptr++] & 0xFF ) << " + (8 * i) + ") | ");
		}
		pw.println("(_msg[ptr++] & 0xFF)));");
		pw.println("\t\t}");
		pw.println("\t\telse");
		pw.println("\t\t{");
		pw.print("\t\t\t_result." + name + " = java.lang.Float.intBitsToFloat((int) (" );
		int i = 0;
		for(i = 0; i < paramSize-1; i++)
		{
			pw.print("((_msg[ptr++] & 0xFF ) << " + (8 * i) + ") | ");
		}
		pw.println("((_msg[ptr++] & 0xFF) << " + (8*i) + " )));");
		pw.println("\t\t}");

	}

	/*********************************************************
	* Prints the un-marshaling of double in an exception
	* 
	* @param PrintWriter The outputstream.
	*
	* @param String The name of the variable
	*********************************************************/
	private void doUnMarshalDouble(PrintWriter pw, String name)
	{
		if(!doubleDef)
			pw.println("\t\tlong dval = 0;");
		if(!longDef)
		{
			pw.println("\t\tlong lval1 = 0;");
			pw.println("\t\tlong lval2 = 0;");
		}
		longDef = true;
		doubleDef = true;		

		int paramSize = codeGenHelper.getParamSize("double");

		if(forceEndian && bigEndian)
		{
			pw.print("\t\tlval1 = (long) (" );
			for(int i = (paramSize/2)-1; i > 0; i--)
			{	
				pw.print("((_msg[ptr++] & 0xFF ) << " + (8 * i) + ") | ");
			}
			pw.println("(_msg[ptr++] & 0xFF));");

			pw.print("\t\tlval2 = (long) (" );
			for(int i = (paramSize/2)-1; i > 0; i--)
			{
				pw.print("((_msg[ptr++] & 0xFF ) << " + (8 * i) + ") | ");
			}
			pw.println("(_msg[ptr++] & 0xFF));");

			pw.println("\t\tdval = (long) ((lval1 << 32) | (lval2 & 0xFFFFFFFFl));");
		}
		else if(forceEndian && !bigEndian)
		{
			pw.print("\t\tlval1 = (long) (" );
			int i = 0;
			for(i = 0; i < (paramSize/2)-1; i++)
			{
				pw.print("((_msg[ptr++] & 0xFF ) << " + (8 * i) + ") | ");
			}
			pw.println("((_msg[ptr++] & 0xFF) << " + (8*i) + " ));");	

			pw.print("\t\tlval2 = (long) (" );
			for(i = 0; i < (paramSize/2)-1; i++)
			{
				pw.print("((_msg[ptr++] & 0xFF ) << " + (8 * i) + ") | ");
			}
			pw.println("((_msg[ptr++] & 0xFF) << " + (8*i) + " ));");	

			pw.println("\t\tdval = (long) ((lval2 << 32) | (lval1 & 0xFFFFFFFFl));");
		}
		else
		{
			pw.println("\t\tif(bigEndian)");
			pw.println("\t\t{");

			pw.print("\t\t\tlval1 = (long) (" );
			for(int i = (paramSize/2)-1; i > 0; i--)
			{	
				pw.print("((_msg[ptr++] & 0xFF ) << " + (8 * i) + ") | ");
			}
			pw.println("(_msg[ptr++] & 0xFF));");

			pw.print("\t\t\tlval2 = (long) (" );
			for(int i = (paramSize/2)-1; i > 0; i--)
			{
				pw.print("((_msg[ptr++] & 0xFF ) << " + (8 * i) + ") | ");
			}
			pw.println("(_msg[ptr++] & 0xFF));");

			pw.println("\t\t\tdval = (long) ((lval1 << 32) | (lval2 & 0xFFFFFFFFl));");
			pw.println("\t\t}");
			pw.println("\t\telse");
			pw.println("\t\t{");
			pw.print("\t\t\tlval1 = (long) (" );
			int i = 0;
			for(i = 0; i < (paramSize/2)-1; i++)
			{
				pw.print("((_msg[ptr++] & 0xFF ) << " + (8 * i) + ") | ");
			}
			pw.println("((_msg[ptr++] & 0xFF) << " + (8*i) + " ));");	

			pw.print("\t\t\tlval2 = (long) (" );
			for(i = 0; i < (paramSize/2)-1; i++)
			{
				pw.print("((_msg[ptr++] & 0xFF ) << " + (8 * i) + ") | ");
			}
			pw.println("((_msg[ptr++] & 0xFF) << " + (8*i) + " ));");	

			pw.println("\t\t\tdval = (long) ((lval2 << 32) | (lval1 & 0xFFFFFFFFl));");

			pw.println("\t\t}");
		}	
		pw.println("\t\t_result." + name + " = (double) java.lang.Double.longBitsToDouble(dval);");
	}

	/*********************************************************
	* Prints the un-marshaling of long in an exception
	* 
	* @param PrintWriter The outputstream.
	*
	* @param String The name of the variable
	*********************************************************/
	private void doUnMarshalLong(PrintWriter pw, String name)
	{
		if(!longDef)
		{
			pw.println("\t\tlong lval1 = 0;");
			pw.println("\t\tlong lval2 = 0;");
		}
		longDef = true;
		
		int paramSize = codeGenHelper.getParamSize("long long");
		if(forceEndian && bigEndian)
		{
			pw.print("\t\tlval1 = (long) (" );
			for(int i = (paramSize/2)-1; i > 0; i--)
			{
				pw.print("((_msg[ptr++] & 0xFF ) << " + (8 * i) + ") | ");
			}
			pw.println("(_msg[ptr++] & 0xFF));");	

			pw.print("\t\tlval2 = (long) (" );
			for(int i = (paramSize/2)-1; i > 0; i--)
			{
				pw.print("((_msg[ptr++] & 0xFF ) << " + (8 * i) + ") | ");
			}
			pw.println("(_msg[ptr++] & 0xFF));");	

			pw.println("\t\t_result." + name + " = (long) ((lval1 << 32) | (lval2 & 0xFFFFFFFFl));");
		}
		else if(forceEndian && !bigEndian)
		{
			pw.print("\t\tlval1 = (long) (" );
			int i = 0;
			for(i = 0; i < (paramSize/2)-1; i++)
			{
				pw.print("((_msg[ptr++] & 0xFF ) << " + (8 * i) + ") | ");
			}
			pw.println("((_msg[ptr++] & 0xFF) << " + (8*i) + " ));");	

			pw.print("\t\tlval2 = (long) (" );
			for(i = 0; i < (paramSize/2)-1; i++)
			{
				pw.print("((_msg[ptr++] & 0xFF ) << " + (8 * i) + ") | ");
			}
			pw.println("((_msg[ptr++] & 0xFF) << " + (8*i) + " ));");	
			pw.println("\t\t_result." + name + " = (long) ((lval2 << 32) | (lval1 & 0xFFFFFFFFl));");
		}
		else
		{
			pw.println("\t\tif(bigEndian)");
			pw.println("\t\t{");

			pw.print("\t\t\tlval1 = (long) (" );
			for(int i = (paramSize/2)-1; i > 0; i--)
			{
				pw.print("((_msg[ptr++] & 0xFF ) << " + (8 * i) + ") | ");
			}
			pw.println("(_msg[ptr++] & 0xFF));");	

			pw.print("\t\t\tlval2 = (long) (" );
			for(int i = (paramSize/2)-1; i > 0; i--)
			{
				pw.print("((_msg[ptr++] & 0xFF ) << " + (8 * i) + ") | ");
			}
			pw.println("(_msg[ptr++] & 0xFF));");	

			pw.println("\t\t_result." + name + " = (long) ((lval1 << 32) | (lval2 & 0xFFFFFFFFl));");
			pw.println("\t\t}");
			pw.println("\t\telse");
			pw.println("\t\t{");
			pw.print("\t\t\tlval1 = (long) (" );
			int i = 0;
			for(i = 0; i < (paramSize/2)-1; i++)
			{
				pw.print("((_msg[ptr++] & 0xFF ) << " + (8 * i) + ") | ");
			}
			pw.println("((_msg[ptr++] & 0xFF) << " + (8*i) + " ));");	

			pw.print("\t\t\tlval2 = (long) (" );
			for(i = 0; i < (paramSize/2)-1; i++)
			{
				pw.print("((_msg[ptr++] & 0xFF ) << " + (8 * i) + ") | ");
			}
			pw.println("((_msg[ptr++] & 0xFF) << " + (8*i) + " ));");	

			pw.println("\t\t_result." + name + " = (long) ((lval2 << 32) | (lval1 & 0xFFFFFFFFl));");
			pw.println("\t\t}");
		}
	}

	/*********************************************************
	* Prints the un-marshaling of boolean in a client
	* 
	* @param PrintWriter The outputstream.
	*********************************************************/
	private void doUnMarshalBoolean(PrintWriter pw)
	{
		pw.println("(_msg[ptr++] == 1) ? true : false);");
	}
	
/**************************************************************************************/

	/*********************************************************
	* Prints the marshaling of float in an exception 
	* Used when MQC-IOP protocol is used.
	
	* @param PrintWriter The outputstream.
	*        
	* @param String Name of argument
	*
	* @param int The current index
	*
	* @return int The new index
	*********************************************************/
	private int doMQCMarshalFloat(PrintWriter pw, String name, int index)
	{
		if(!floatDef)
			pw.print("\t\tint fval = ");
		else
			pw.print("\t\tfval = ");
		floatDef = true;
		
		pw.println("java.lang.Float.floatToIntBits(value." + name + ");");		

		int paramSize = codeGenHelper.getParamSize("float");
		if(forceEndian && !bigEndian)
		{
			for(int i = 0; i < paramSize-1; i++)
			{
				pw.print("\t\t_msg[" + index++ + "] = (byte) ");
				pw.println("(fval >> " + (8 * i) + ");" );
			}
			pw.println("\t\t_msg[" + index++ + "] = (byte) (fval >> " + (8 * (paramSize-1)) + ");" );
		}
		else
		{
			for(int i = paramSize-1; i > 0; i--)
			{
				pw.print("\t\t_msg[" + index++ + "] = (byte) ");
				pw.println("(fval >> " + (8 * i) + ");" );
			}
			pw.println("\t\t_msg[" + index++ + "] = (byte) fval;");
		}
		return index;
	}


	/*********************************************************
	* Prints the marshaling of double in an exception
	* 
	* @param PrintWriter The outputstream.
	*        
	* @param String The name of the argument
	*
	* @param int The current index
	*
	* @return int The new index
	*********************************************************/
	private int doMQCMarshalDouble(PrintWriter pw, String name, int index)
	{
		if(!doubleDef)
			pw.print("\t\tlong dval = ");
		else
			pw.print("\t\tdval = ");
		doubleDef = true;
		
		pw.println("java.lang.Double.doubleToLongBits(value." + name + ");");

		int paramSize = codeGenHelper.getParamSize("double");
		if(forceEndian && !bigEndian)
		{
			for(int i = 0; i < paramSize-1; i++)
			{
				pw.print("\t\t_msg[" + index++ + "] = (byte) ");
				pw.println("(dval >> " + (8 * i) + ");" );
			}
			pw.println("\t\t_msg[" + index++ + "] = (byte) (dval >> " + (8 * (paramSize-1)) + ");" );
		}
		else
		{
			for(int i = paramSize-1; i > 0; i--)
			{
				pw.print("\t\t_msg[" + index++ + "] = (byte) ");
				pw.println("(dval >> " + (8 * i) + ");" );
			}
			pw.println("\t\t_msg[" + index++ + "] = (byte) dval;");
		}
		return index;		
	}

	/*********************************************************
	* Prints the un-marshaling of float in a exception
	* 
	* @param PrintWriter The outputstream.
	*
	* @param String Name of param
	*
	* @param int The current index
	*
	* @return int The new index
	*********************************************************/
	private int doMQCUnMarshalFloat(PrintWriter pw, String name, int index)
	{
		int paramSize = codeGenHelper.getParamSize("float");
		int tmpIndex = index;
		
		pw.println("\t\tif(bigEndian)");
		pw.println("\t\t{");

		pw.print("\t\t\t_result." + name + " = java.lang.Float.intBitsToFloat((int) (" );
		for(int i = paramSize-1; i > 0; i--)
		{
			pw.print("((_msg[" + tmpIndex++ + "] & 0xFF ) << " + (8 * i) + ") | ");
		}
		pw.println("(_msg[" + tmpIndex++ + "] & 0xFF)));");
		pw.println("\t\t}");
		pw.println("\t\telse");
		pw.println("\t\t{");
		pw.print("\t\t\t_result." + name + " = java.lang.Float.intBitsToFloat((int) (" );
		int i = 0;
		for(i = 0; i < paramSize-1; i++)
		{
			pw.print("((_msg[" + index++ + "] & 0xFF ) << " + (8 * i) + ") | ");
		}
		pw.println("((_msg[" + index++ + "] & 0xFF) << " + (8*i) + " )));");
		pw.println("\t\t}");
		return index;
	}

	/*********************************************************
	* Prints the un-marshaling of double in an exception
	* 
	* @param PrintWriter The outputstream.
	*
	* @param String The name of the variable
	*
	* @param int The current index
	*
	* @return int The new index
	*********************************************************/
	private int doMQCUnMarshalDouble(PrintWriter pw, String name, int index)
	{
		if(!doubleDef)
			pw.println("\t\tlong dval = 0;");
		if(!longDef)
		{
			pw.println("\t\tlong lval1 = 0;");
			pw.println("\t\tlong lval2 = 0;");
		}
		longDef = true;
		doubleDef = true;		
		int tmpIndex = index;
		
		int paramSize = codeGenHelper.getParamSize("double");

		if(forceEndian && bigEndian)
		{
			pw.print("\t\tlval1 = (long) (" );
			for(int i = (paramSize/2)-1; i > 0; i--)
			{	
				pw.print("((_msg[" + index++ + "] & 0xFF ) << " + (8 * i) + ") | ");
			}
			pw.println("(_msg[" + index++ + "] & 0xFF));");

			pw.print("\t\tlval2 = (long) (" );
			for(int i = (paramSize/2)-1; i > 0; i--)
			{
				pw.print("((_msg[" + index++ + "] & 0xFF ) << " + (8 * i) + ") | ");
			}
			pw.println("(_msg[" + index++ + "] & 0xFF));");

			pw.println("\t\tdval = (long) ((lval1 << 32) | (lval2 & 0xFFFFFFFFl));");
		}
		else if(forceEndian && !bigEndian)
		{
			pw.print("\t\tlval1 = (long) (" );
			int i = 0;
			for(i = 0; i < (paramSize/2)-1; i++)
			{
				pw.print("((_msg[" + index++ + "] & 0xFF ) << " + (8 * i) + ") | ");
			}
			pw.println("((_msg[" + index++ + "] & 0xFF) << " + (8*i) + " ));");	

			pw.print("\t\tlval2 = (long) (" );
			for(i = 0; i < (paramSize/2)-1; i++)
			{
				pw.print("((_msg[" + index++ + "] & 0xFF ) << " + (8 * i) + ") | ");
			}
			pw.println("((_msg[" + index++ + "] & 0xFF) << " + (8*i) + " ));");	

			pw.println("\t\tdval = (long) ((lval2 << 32) | (lval1 & 0xFFFFFFFFl));");
		}
		else
		{
			pw.println("\t\tif(bigEndian)");
			pw.println("\t\t{");

			pw.print("\t\t\tlval1 = (long) (" );
			for(int i = (paramSize/2)-1; i > 0; i--)
			{	
				pw.print("((_msg[" + index++ + "] & 0xFF ) << " + (8 * i) + ") | ");
			}
			pw.println("(_msg[" + index++ + "] & 0xFF));");

			pw.print("\t\t\tlval2 = (long) (" );
			for(int i = (paramSize/2)-1; i > 0; i--)
			{
				pw.print("((_msg[" + index++ + "] & 0xFF ) << " + (8 * i) + ") | ");
			}
			pw.println("(_msg[" + index++ + "] & 0xFF));");

			pw.println("\t\t\tdval = (long) ((lval1 << 32) | (lval2 & 0xFFFFFFFFl));");
			pw.println("\t\t}");
			pw.println("\t\telse");
			pw.println("\t\t{");
			pw.print("\t\t\tlval1 = (long) (" );
			int i = 0;
			for(i = 0; i < (paramSize/2)-1; i++)
			{
				pw.print("((_msg[" + tmpIndex++ + "] & 0xFF ) << " + (8 * i) + ") | ");
			}
			pw.println("((_msg[" + tmpIndex++ + "] & 0xFF) << " + (8*i) + " ));");	

			pw.print("\t\t\tlval2 = (long) (" );
			for(i = 0; i < (paramSize/2)-1; i++)
			{
				pw.print("((_msg[" + tmpIndex++ + "] & 0xFF ) << " + (8 * i) + ") | ");
			}
			pw.println("((_msg[" + tmpIndex++ + "] & 0xFF) << " + (8*i) + " ));");	

			pw.println("\t\t\tdval = (long) ((lval2 << 32) | (lval1 & 0xFFFFFFFFl));");

			pw.println("\t\t}");
		}	
		pw.println("\t\t_result." + name + " = (double) java.lang.Double.longBitsToDouble(dval);");
		return index;
	}

	/*********************************************************
	* Prints the un-marshaling of long in an exception
	* 
	* @param PrintWriter The outputstream.
	*
	* @param String The name of the variable
	*
	* @param int The current index
	*
	* @return int The new index
	*********************************************************/
	private int doMQCUnMarshalLong(PrintWriter pw, String name, int index)
	{
		if(!longDef)
		{
			pw.println("\t\tlong lval1 = 0;");
			pw.println("\t\tlong lval2 = 0;");
		}
		longDef = true;
		
		int tmpIndex = index;
		
		int paramSize = codeGenHelper.getParamSize("long long");
		if(forceEndian && bigEndian)
		{
			pw.print("\t\tlval1 = (long) (" );
			for(int i = (paramSize/2)-1; i > 0; i--)
			{
				pw.print("((_msg[" + index++ + "] & 0xFF ) << " + (8 * i) + ") | ");
			}
			pw.println("(_msg[" + index++ + "] & 0xFF));");	

			pw.print("\t\tlval2 = (long) (" );
			for(int i = (paramSize/2)-1; i > 0; i--)
			{
				pw.print("((_msg[" + index++ + "] & 0xFF ) << " + (8 * i) + ") | ");
			}
			pw.println("(_msg[" + index++ + "] & 0xFF));");	

			pw.println("\t\t_result." + name + " = (long) ((lval1 << 32) | (lval2 & 0xFFFFFFFFl));");
		}
		else if(forceEndian && !bigEndian)
		{
			pw.print("\t\tlval1 = (long) (" );
			int i = 0;
			for(i = 0; i < (paramSize/2)-1; i++)
			{
				pw.print("((_msg[" + index++ + "] & 0xFF ) << " + (8 * i) + ") | ");
			}
			pw.println("((_msg[" + index++ + "] & 0xFF) << " + (8*i) + " ));");	

			pw.print("\t\tlval2 = (long) (" );
			for(i = 0; i < (paramSize/2)-1; i++)
			{
				pw.print("((_msg[" + index++ + "] & 0xFF ) << " + (8 * i) + ") | ");
			}
			pw.println("((_msg[" + index++ + "] & 0xFF) << " + (8*i) + " ));");	
			pw.println("\t\t_result." + name + " = (long) ((lval2 << 32) | (lval1 & 0xFFFFFFFFl));");
		}
		else
		{
			pw.println("\t\tif(bigEndian)");
			pw.println("\t\t{");

			pw.print("\t\t\tlval1 = (long) (" );
			for(int i = (paramSize/2)-1; i > 0; i--)
			{
				pw.print("((_msg[" + index++ + "] & 0xFF ) << " + (8 * i) + ") | ");
			}
			pw.println("(_msg[" + index++ + "] & 0xFF));");	

			pw.print("\t\t\tlval2 = (long) (" );
			for(int i = (paramSize/2)-1; i > 0; i--)
			{
				pw.print("((_msg[" + index++ + "] & 0xFF ) << " + (8 * i) + ") | ");
			}
			pw.println("(_msg[" + index++ + "] & 0xFF));");	

			pw.println("\t\t_result." + name + " = (long) ((lval1 << 32) | (lval2 & 0xFFFFFFFFl));");
			pw.println("\t\t}");
			pw.println("\t\telse");
			pw.println("\t\t{");
			pw.print("\t\t\tlval1 = (long) (" );
			int i = 0;
			for(i = 0; i < (paramSize/2)-1; i++)
			{
				pw.print("((_msg[" + tmpIndex++ + "] & 0xFF ) << " + (8 * i) + ") | ");
			}
			pw.println("((_msg[" + tmpIndex++ + "] & 0xFF) << " + (8*i) + " ));");	

			pw.print("\t\t\tlval2 = (long) (" );
			for(i = 0; i < (paramSize/2)-1; i++)
			{
				pw.print("((_msg[" + tmpIndex++ + "] & 0xFF ) << " + (8 * i) + ") | ");
			}
			pw.println("((_msg[" + tmpIndex++ + "] & 0xFF) << " + (8*i) + " ));");	

			pw.println("\t\t_result." + name + " = (long) ((lval2 << 32) | (lval1 & 0xFFFFFFFFl));");
			pw.println("\t\t}");
		}
		return index;
	}

	/*********************************************************
	* Prints the un-marshaling of boolean in a client
	* 
	* @param PrintWriter The outputstream.
	*
	* @param int The current index
	*
	* @return int The new index
	*********************************************************/
	private int doMQCUnMarshalBoolean(PrintWriter pw, int index)
	{
		pw.println("(_msg[" + index++ + "] == 1) ? true : false);");
		return index;
	}

}
