/*
 * Copyright (c) 2003 David E. Bakken, his research students, and Washington State University.
 * Please see the file LICENSE.pdf for more details on terms of use.
 */

package mqc.ior;

import java.lang.*;
import java.util.*;

/**
* Class for an IOR 
*
* Initial IOR class as implemented by Olav Haugan and then
* extended with support for corbaloc references by David
* McKinnon
*
* @author Olav Haugan
* @author A. David McKinnon
*/
public class IOR
{
   private String  IORStr;
   private String  corbalocStr;
   private Profile profile;
   private byte[]  objectKey;

   private IORParser parser;

   /**
   * Constructor. Initializes variables
   *
   */
   public IOR()
   {
      IORStr = null;
      corbalocStr = null;
      parser = new IORParser();
   }

   /**
   * Constructor. Initializes variables
   *
   */
   public IOR(Profile profile)
   {
      IORStr       = null;
      corbalocStr  = null;
      this.profile = profile;
      parser       = new IORParser();
   }

   /**
   * Constructor. Initializes variables
   *
   * @param String The IOR string
   */
   public IOR(String IORStr)
   {
      this.IORStr      = IORStr;
      this.corbalocStr = null;
      parser           = new IORParser();
   }

   /**
   * Constructor. Initializes variables
   *
   * @param String The IOR string
   * @param String The corbaloc string
   */
   public IOR(String IORStr, String corbalocStr)
   {
      this.IORStr      = IORStr;
      this.corbalocStr = corbalocStr;
      parser           = new IORParser();
   }

   /**
   * Parses the IOR string and gets the profile
   *
   * @return boolean true if string is valid, false otherwise
   */
   public boolean decode()
   {
      try
      {
         if(parser.parseIOR(IORStr))
         {
            profile = parser.getProfile();
            return true;
         }
         else
            return false;
      }
      catch ( RuntimeException re ) 
      {
         /* Hack!! */
         /* If we get a RuntimeException this probably means that 
          * we do NOT support the tag that was parsed. For now, 
          * lets get the profile object anyway in the case where
          * a "good" profile tag was found before the exception
          * was raised. */
         profile = parser.getProfile();
         return true;
      }
   }

   /**
   * Parses a corbaloc string and gets the profile
   *
   * @return boolean true if string is valid, false otherwise
   */
   public boolean decodeCorbaloc()
   {
      String temp;
      profile = new Profile();
      Enumeration e;
      int start = 0;
      int end = 0;
      int offset = 0;

      temp = corbalocStr.substring(0, 9);
      if (!temp.equals("corbaloc:"))
      {
         System.out.println("Unknown Corba Location format " +
                            "corbaloc:<protocol>:<host>:<port>/Object/key;<protocol>:<host>....\n" + corbalocStr);
         return false;
      }
      else
      {
         start = 9;
      }

      /* we have several occurences of corbalocs */
      while ((end = corbalocStr.substring(start+end+1, corbalocStr.length()).indexOf(",")) != -1)
      {
         offset = 1;
         temp = corbalocStr.substring(start, start + end +1);       /* Get the corbaloc string */
         if (parser.parseCorbaloc("corbaloc:" + temp)) /* Make sure we have an element */
         {
            /* get the result from the parser and get the elements */
            e = parser.getProfile().getProfiles();

            /* We expect only one element */
            profile.add((TaggedProfile) e.nextElement());
            start = start + end + 1;
         } else
         {
            return false;
         }
      }

      /* Okay, now for the last element */
      if(parser.parseCorbaloc("corbaloc:" + corbalocStr.substring(start+offset, corbalocStr.length())))
      {
         e = parser.getProfile().getProfiles();
         profile.add((TaggedProfile) e.nextElement());
         return true;
      }
      else
      {
         return false;
      }
   }

   /**
   * Returns the profile
   *
   * @return Profile profile
   */
   public Profile getProfile()
   {
      return profile;
   }

   /**
   * Returns the object key
   *
   * @return byte[] The object key
   */
   public byte[] getObjectKey()
   {
      Enumeration e = profile.taggedProfile.elements();
      TaggedProfile tagProfile = null;

      if(e.hasMoreElements())
      {
         tagProfile = (TaggedProfile) e.nextElement();
         return tagProfile.objectKey;
      }
      return null;
   }

   /**
   * Sets the profile
   *
   * @param Profile profile
   */
   public void setProfile(Profile profile)
   {
      this.profile = profile;
   }

   /**
   * Encodes the profile to an IOR string
   */
   public boolean encode()
   {
      IORStr = parser.encodeProfile(profile);
      return true;
   }

   /**
   * Encodes the profile to a corbaloc string
   */
   public boolean encodeCorbaloc()
   {
      corbalocStr = parser.encodeCorbalocProfile(profile);
      return true;
   }

   /**
   * Returns the IOR string
   *
   * @return String The IOR string
   */
   public String getIORStr()
   {
      return IORStr;
   }

   /**
   * Returns the corbaloc string
   *
   * @return String The corbaloc string
   */
   public String getCorbalocStr()
   {
      return corbalocStr;
   }
}
