/*
 * Copyright (c) 2003 David E. Bakken, his research students, and Washington State University.
 * Please see the file LICENSE.pdf for more details on terms of use.
 */

/* Null256.java
 *
 * Adapted from the Cryptix Null MessageDigest file by David McKinnon.
 */

package mqc.security.md;

import mqc.security.MessageDigest;
import mqc.security.DigestException;

/**
 * Implements a Null message digest algorithm in Java.
 *
 * @author  David McKinnon
 */
public final class Null256 extends MessageDigest implements Cloneable
{

   // Static variables and constants
   //...........................................................................

   /** The size in bytes of the input block to the tranformation algorithm. */
   private static final int BLOCK_SIZE = 32;
   private static final int HASH_SIZE  = 32;     // 256-bit

   // Instance variables
   //...........................................................................

   /** Digest Buffer */
   private final byte[] buf;

   // Constructors, clone()
   //...........................................................................

   public Null256 ()
   {
      buf = new byte[HASH_SIZE];
   }

   /** This constructor is here to implement cloneability of this class. */
   private Null256(Null256 md)
   {
      // no state to clone, just need to create a new buffer
      buf = new byte[HASH_SIZE];
   }

   /** Returns a copy of this MD object. */
   public Object clone()
   {
      return new Null256(this);
   }

   // JCE methods
   //...........................................................................

   /** returns the BLOCK_SIZE */
   protected int engineGetBlockSize()
   {
      return BLOCK_SIZE;
   }

   /** returns the HASH_SIZE */
   protected int engineGetDigestLength()
   {
      return HASH_SIZE;
   }

   /**
    * Resets this object disregarding any temporary data present at the
    * time of the invocation of this call.
    */
   protected void engineReset()
   {
      // nothing to reset
   }

   /**
    * Continues a Null message digest using the input byte.
    */
   public void engineUpdate (byte input)
   {
      // nothing to update
   }

   /**
    * Null block update operation.
    */
   protected void engineUpdate (byte[] input, int offset, int len)
   {
      // nothing to do
   }

   /**
    * Completes the Null hash computation
    */
   protected byte[] engineDigest ()
   {
      byte[] tmp = new byte[HASH_SIZE];
      System.arraycopy(buf, 0, tmp, 0, HASH_SIZE);
      return tmp;
   }

   /**
    * Completes the Null hash computation
    */
   protected int engineDigest(byte[] buf, int offset, int len)
      throws DigestException
   {
      if (len < HASH_SIZE)
      throw new DigestException();

      System.arraycopy(this.buf, 0, buf, offset, HASH_SIZE);

      return HASH_SIZE;
   }

   // own methods
   //...........................................................................

   // none, n/a
}
