/*
 * Copyright (c) 2003 David E. Bakken, his research students, and Washington State University.
 * Please see the file LICENSE.pdf for more details on terms of use.
 */

/* $Id: Parity8.java,v 1.2 2003/09/08 18:53:04 rjohnsto Exp $
 *
 * Perform a simply digest based upon a running byte-by-byte sum (Byte Parity).
 */

package mqc.security.md;

import mqc.security.MessageDigest;
import mqc.security.DigestException;

/**
 * Implements a Parity8-based message digest algorithm in Java.
 *
 * @author  A. David McKinnon
 */
public final class Parity8 extends MessageDigest implements Cloneable
{

   // Static variables and constants
   //...........................................................................

   /** The size in bytes of the input block to the tranformation algorithm. */
   private static final int BLOCK_LENGTH = 1; //    = 8 / 8;


   // Instance variables
   //...........................................................................

   /** temporary internal parity value */
   private byte parity;    // currently calculated parity


   // Constructors, clone()
   //...........................................................................

   public Parity8 ()
   {
      super();
      engineReset();
   }

   /** This constructor is here to implement cloneability of this class. */
   private Parity8(Parity8 md)
   {
      this();
   }

   /** Returns a copy of this MD object. */
   public Object clone()
   {
      return new Parity8(this);
   }


   // JCE methods
   //...........................................................................

   /** returns the Digest Length (BLOCK_LENGTH) */
   protected int engineGetDigestLength()
   {
      return BLOCK_LENGTH;
   }

   /** returns the BLOCK_LENGTH */
   protected int engineGetBlockSize()
   {
      return BLOCK_LENGTH;
   }

   /**
    * Resets this object disregarding any temporary data present at the
    * time of the invocation of this call.
    */
   protected void engineReset()
   {
      parity = 0;
   }

   /**
    * Continues a Parity8 message digest using the input byte.
    */
   public void engineUpdate (byte input)
   {
      parity += input;
   }


   /**
    * Parity8 block update operation.
    * <p>
    * Continues an Parity8 message digest operation.
    *
    * @param    input    input block
    * @param    offset   start of meaningful bytes in input
    * @param    len      count of bytes in input block to consider
    */
   protected void engineUpdate (byte[] input, int offset, int len)
   {
      // make sure we don't exceed input's allocated size/length
      if (offset < 0 || len < 0 || (long)offset + len > input.length)
         throw new ArrayIndexOutOfBoundsException();

      int i = offset;

      while (len-- != 0)
      {
         parity += input[i++];
      }
   }

   /**
    * Completes the hash computation by performing final operations such
    * as padding. At the return of this engineDigest, the MD engine is
    * reset.
    *
    * @return    the array of bytes for the resulting hash value.
    */
   protected byte[] engineDigest ()
   {
      // copy the long into a 4-byte array
      byte[] result = new byte[BLOCK_LENGTH];

      // write the 1-byte parity value
      result[0] = parity;

      // reset the engine
      engineReset();
      return result;
   }

   /**
    * Calls engineDigest to do the actual work
    */
   protected int engineDigest(byte[] buf, int offset, int len)
   throws DigestException
   {
      if (len < BLOCK_LENGTH)
         throw new DigestException();

      byte [] eDigest = engineDigest();
      System.arraycopy( eDigest, 0, buf, offset, eDigest.length);

      return eDigest.length;
   }


   // own methods
   //...........................................................................

   // not needed, the Parity8 code is embedded in the above method bodies
}
