/*
 * Copyright (c) 2003 David E. Bakken, his research students, and Washington State University.
 * Please see the file LICENSE.pdf for more details on terms of use.
 */

//Use MQC package only to get to work with two different
//libraries
package mqc;

/**
* Class for CDR output stream. Marshals the data to
* the buffer. Everything is marshalled in big endian.
* 
* @author Olav Haugan
*
* 14.10.01
*/
public class CDROutputStream
{
   private boolean bigEndian;
   private byte[] buffer;
   private int bufferPtr;

   /**
   * Constructor. Initializes variables.
   */
   public CDROutputStream()
   {
      buffer = null;
      bufferPtr = 0;
      bigEndian = true;
   }

   /**
   * Constructor. Initializes variables.
   *
   * @param byte[] The buffer to stream to
   * @param int The buffer pointer
   */
   public CDROutputStream(byte[] buffer, int bufferPtr)
   {
      this.buffer = buffer;
      this.bufferPtr = bufferPtr;
      this.bigEndian = true;
   }

   /**
   * Constructor. Initializes variables.
   *
   * @param byte[] The buffer to stream to
   * @param int The buffer pointer
   * @param boolean true if bigendian, false otherwise
   */
   public CDROutputStream(byte[] buffer, int bufferPtr, boolean bigEndian)
   {
      this.buffer = buffer;
      this.bufferPtr = bufferPtr;
      this.bigEndian = bigEndian;
   }

   /**
   * Marshals a byte to the stream.
   *
   * @param byte a byte
   */
   public void write_octet(byte octet)
   {
      bufferResizer(1);
      buffer[bufferPtr++] = octet;
   }

   /**
   * Marshals a boolean to the stream.
   *
   * @param Boolean A boolean
   */
   public void write_boolean(boolean value)
   {
      bufferResizer(1);
      buffer[bufferPtr++] = (byte) (value ? 1 : 0);
   }

   /**
   * Marshals a char to the stream.
   *
   * @param char A char
   */
   public void write_char(char value)
   {
      bufferResizer(1);
      buffer[bufferPtr++] = (byte) value;
   }

   /**
   * Marshals a unsigned short to the stream.
   *
   * @param unsigned short An unsigned short
   */
   public void write_ushort(short value)
   {
      write_short(value);
   }

   /**
   * Marshals a short to the stream.
   *
   * @param short A short
   */
   public void write_short(short value)
   {
      //Align
      bufferPtr = (bufferPtr +1) & ~1;
      bufferResizer(2);
      if(bigEndian)
      {
         buffer[bufferPtr++] = (byte) (value >> 8);
         buffer[bufferPtr++] = (byte) value;
      }
      else
      {
         buffer[bufferPtr++] = (byte) (value);
         buffer[bufferPtr++] = (byte) (value >> 8);
      }
   }

   /**
   * Marshals a unsigned long to the stream.
   *
   * @param unsigned long An unsigned long
   */
   public void write_ulong(int value)
   {
      write_long(value);
   }

   /**
   * Marshals a long to the stream.
   *
   * @param long A long
   */
   public void write_long(int value)
   {
      //Align
      bufferPtr = (bufferPtr +3) & ~3;
      bufferResizer(4);
      if(bigEndian)
      {
         buffer[bufferPtr++] = (byte) (value >> 24);
         buffer[bufferPtr++] = (byte) (value >> 16);
         buffer[bufferPtr++] = (byte) (value >> 8);
         buffer[bufferPtr++] = (byte) value;
      }
      else
      {
         buffer[bufferPtr++] = (byte) (value >> 0);
         buffer[bufferPtr++] = (byte) (value >> 8);
         buffer[bufferPtr++] = (byte) (value >> 16);
         buffer[bufferPtr++] = (byte) (value >> 24);
      }
   }

   /**
   * Marshals an unsigned long long to the stream.
   *
   * @param unsigned long long An unsigned long long
   */
   public void write_ulonglong(long value)
   {
      write_longlong(value);
   }

   /**
   * Marshals a long long to the stream.
   *
   * @param long long A long long
   */
   public void write_longlong(long value)
   {
      //Align
      bufferPtr = (bufferPtr +7) & ~7;
      bufferResizer(8);
      if(bigEndian)
      {
         buffer[bufferPtr++] = (byte) (value >> 56);
         buffer[bufferPtr++] = (byte) (value >> 48);
         buffer[bufferPtr++] = (byte) (value >> 40);
         buffer[bufferPtr++] = (byte) (value >> 32);
         buffer[bufferPtr++] = (byte) (value >> 24);
         buffer[bufferPtr++] = (byte) (value >> 16);
         buffer[bufferPtr++] = (byte) (value >>  8);
         buffer[bufferPtr++] = (byte) value;
      }
      else
      {
         buffer[bufferPtr++] = (byte) (value >>  0);
         buffer[bufferPtr++] = (byte) (value >>  8);
         buffer[bufferPtr++] = (byte) (value >> 16);
         buffer[bufferPtr++] = (byte) (value >> 24);
         buffer[bufferPtr++] = (byte) (value >> 32);
         buffer[bufferPtr++] = (byte) (value >> 40);
         buffer[bufferPtr++] = (byte) (value >> 48);
         buffer[bufferPtr++] = (byte) (value >> 56);
      }
   }

   /**
   * Marshals a float to the stream.
   *
   * @param float A float
   */
   public void write_float(float value)
   {
      write_long(java.lang.Float.floatToIntBits(value));
   }

   /**
   * Marshals a double to the stream.
   *
   * @param double A double
   */
   public void write_double(double value)
   {
      write_longlong(java.lang.Double.doubleToLongBits(value));
   }

   /**
   * Marshals a string to the stream.
   *
   * @param String A string
   */
   public void write_string(String value)
   {
      int len = value.length() + 1;
      write_long(len);
      bufferResizer(len);
      System.arraycopy(value.getBytes(), 0, buffer, bufferPtr, len-1);
      bufferPtr += len-1;
      write_octet((byte)0);
   }

   /**
   * Resizes the buffer if it is too small
   *
   * @param int The size of the value that are being marshaled
   */
   private void bufferResizer(int size)
   {
      if(buffer.length < (bufferPtr + size))
      {
         byte[] tmp = new byte[buffer.length + size + 10];
         System.arraycopy(buffer, 0, tmp, 0, buffer.length);
         buffer = tmp;
      }
   }

   /**
   * Gets the pointer to the buffer
   *
   * @return int The buffer pointer
   */
   public int getBufferPtr()
   {
      return bufferPtr;
   }

   /**
   * Gets the buffer
   *
   * @return byte[] The buffer
   *
   */
   public byte[] getBuffer()
   {
      return buffer;
   }

   /**
   * Sets the buffer
   *
   * @param byte[] The buffer
   *
   */
   public void setBuffer(byte[] buffer)
   {
      this.buffer = buffer;
   }
}

